import { chain, constant, Self, Value } from './common.js'

/** @import { Construct, Morphism, Set } from './common.js' */

/**
 * @template T, E
 * @typedef {Ok<T> | Err<E>} Result
 */

/**
 * @template T
 * @typedef OkMethods
 * @property {typeof isOk} isOk
 * @property {typeof isErr} isErr
 * @property {typeof chain} chain
 * @property {typeof map} map
 * @property {typeof alt} alt
 * @property {typeof fold} fold
 * @property {typeof bimap} bimap
 */

/**
 * @template T
 * @typedef {Set<T> & Construct<T> & OkMethods<T>} Ok
 */

/**
 * @template T
 * @typedef ErrMethods
 * @property {typeof isOk} isOk
 * @property {typeof isErr} isErr
 * @property {typeof chain} chain
 * @property {typeof map} map
 * @property {typeof alt} alt
 * @property {typeof fold} fold
 * @property {typeof bimap} bimap
 */

/**
 * @template T
 * @typedef {Set<T> & Construct<T> & ErrMethods<T>} Err
 */

/**
 * @template T, U, E
 * @this {Result<T, E>}
 * @param {Morphism<T, U>} fn 
 * @returns {Result<U, E>}
 */
function map(fn) {
  return /** @type {Result<U, E>} */ (this[Self](this.chain(fn)))
}

/**
 * @template T1, T2, E1, E2
 * @this {Result<T1, E1>}
 * @param {Morphism<T1, T2>} y
 * @param {Morphism<E1, E2>} x
 * @returns {Result<T2, E2>}
 */
function bimap(x, y) {
  return map.call(this, y).map(x)
}

/**
 * @template T, U, E
 * @this {Result<T, E>}
 * @param {Morphism<T, U>} fn 
 * @param {U} acc 
 * @return {U}
 */
function fold(fn, acc) {
  const result = this.map(fn)
  return result.isOk() ? result[Value] : acc
}

/**
 * @template T, E
 * @this {Result<T, E>}
 * @param {Result<T, E>} other 
 * @returns {Result<T, E>}
 */
function alt(other) {
  return this.isOk() ? this : other
}

/**
 * @template T, E
 * @this {Result<T, E>}
 * @returns {this is Ok<T>}
 */
function isOk() {
  return this[Self] === Ok
}

/**
 * @template T, E
 * @this {Result<T, E>}
 * @returns {this is Err<E>}
 */
function isErr() {
  return this[Self] === Err
}

/**
 * @template T
 * @param {?T} value
 * @returns {Ok<T>}
 */
export const Ok = value => Object.freeze({
  [Self]: Ok,
  [Value]: value,
  isOk,
  isErr,
  chain,
  map,
  alt,
  fold,
  bimap
})

/**
 * @template T
 * @param {T?} value
 * @returns {Err<T>}
 */
export const Err = value => Object.freeze({
  [Self]: Err,
  [Value]: value,
  chain: constant,
  map: constant,
  alt,
  isOk,
  isErr,
  fold,
  bimap
})

/**
 * @template T, E
 * @param {T} value
 * @returns {Result<T, E>}
 */
export const Result = value => Ok(value)
Result.of = Result
Result.zero = () => Err(undefined)

